#include <stdlib.h>
#include <time.h>
#include <iostream>
#include "object_model.hpp"

template <class T, class VB>
inline bool
is_partial(T *t)
{
  size_t vb_offset = (size_t)static_cast<VB*>(t) - (size_t)t;
  size_t t_size = vb_offset + sizeof(VB);

  return t_size > sizeof *t;
}

template <class T, class VB1, class VB2>
inline bool
is_partial(T *t)
{
  size_t vb_offset1 = (size_t)static_cast<VB1*>(t) - (size_t)t;
  size_t vb_offset2 = (size_t)static_cast<VB2*>(t) - (size_t)t;

  if (vb_offset1 > vb_offset2)
    std::swap(vb_offset1, vb_offset2);

  size_t t_size1 = vb_offset1 + sizeof(VB1) + sizeof(VB2);
  size_t t_size2 = vb_offset2 + sizeof(VB2);

  return t_size1 > sizeof *t || t_size2 > sizeof *t;
}

/*
  ===============================
*/

struct A {
  A(long a_) : a(a_) {
    print_layout("A", "ctor", this, sizeof *this);
  }

  virtual ~A() {
    a = 0;
    print_layout("A", "dtor", this, sizeof *this);
  };

  virtual void f() const {
    print("f");
  }

protected:
  long a;

private:
  void print(const char *func) const {
    std::cout << "A     ::" << func << '('
	      << reinterpret_cast<size_t>(this) << ")\t"
	      << "a = " << a << "\n";
  }
};

struct B {
  B(long b_) : b(b_) {
    print_layout("B", "ctor", this, sizeof *this);
  }

  virtual ~B() {
    b = 0;
    print_layout("B", "dtor", this, sizeof *this);
  }

  virtual B* g() {
    print("g");
    return this;
  }
  
protected:
  long b;
  
private:
  void print(const char *func) const {
    std::cout << "B     ::" << func << '('
	      << reinterpret_cast<size_t>(this) << ")\t"
	      << "b = " << b << "\n";
  }
};

struct C : virtual B {
  C(long b, long c_) : B(b), c(c_) {
    if (is_partial<C,B>(this))
      print_layout("C", "ctor_", this, sizeof *this - sizeof(B));
    else
      print_layout("C", "ctor", this, sizeof *this);
  }

  virtual ~C() {
    c = 0;
    if (is_partial<C,B>(this))
      print_layout("C", "dtor_", this, sizeof *this - sizeof(B));
    else
      print_layout("C", "dtor", this, sizeof *this);
  }
  
  B* g() { // override B::g
    print("g");
    return this;
  }
  
protected:
  long c;
  
private:
  void print(const char *func) const {
    std::cout << "C     ::" << func << '('
	      << reinterpret_cast<size_t>(this) << ")\t"
	      << "b = " << b
	      << ", c = " << c << "\n";
  }
};

struct AB : virtual A, virtual B {
  AB(long a, long b, long ab_) : A(a), B(b), ab(ab_) {
    if (is_partial<AB,A,B>(this))
      print_layout("AB", "ctor_",
		   this, sizeof *this - sizeof(A)-sizeof(B));
    else
      print_layout("AB", "ctor", this, sizeof *this);
  }

  virtual ~AB() {
    ab = 0;
    if (is_partial<AB,A,B>(this))
      print_layout("AB", "dtor_",
		   this, sizeof *this - sizeof(A)-sizeof(B));
    else
      print_layout("AB", "dtor", this, sizeof *this);
  }

  AB* g () { // override B::g
    print("g");
    return this;
  }

  virtual AB* h() {
    print("h");
    return this;
  }

protected:
  long ab;

private:
  void print(const char *func) const {
    std::cout << "AB    ::" << func << '('
	      << reinterpret_cast<size_t>(this) << ")\t"
	      << "a = " << a
	      << ", b = " << b
	      << ", ab = " << ab << "\n";
  }
};

struct BA : virtual B, virtual A {
  BA(long a, long b, long ba_) : B(b), A(a), ba(ba_) {
    if (is_partial<BA,A,B>(this))
      print_layout("BA", "ctor_", this,
		   sizeof *this - sizeof(A)-sizeof(B));
    else
      print_layout("BA", "ctor", this, sizeof *this);
  }

  virtual ~BA() {
    ba = 0;
    if (is_partial<BA,A,B>(this))
      print_layout("BA", "dtor_",
		   this, sizeof *this - sizeof(A)-sizeof(B));
    else
      print_layout("BA", "dtor", this, sizeof *this);
  }

  void f () const { // override A::f
    print("f");
  }
  
protected:
  long ba;

private:
  void print(const char *func) const {
    std::cout << "BA    ::" << func << '('
	      << reinterpret_cast<size_t>(this) << ")\t"
	      << "a = " << a
	      << ", b = " << b
	      << ", ba = " << ba << "\n";
  }
};

struct ABBA : protected AB, BA {
  ABBA(long a, long b, long ab, long ba, long abba_) :
    A(a), B(b), AB(a,b,ab), BA(a,b,ba), abba(abba_) {
    if (is_partial<ABBA,A,B>(this))
      print_layout("ABBA", "ctor_",
		   this, sizeof *this - sizeof(A)-sizeof(B));
    else
      print_layout("ABBA", "ctor", this, sizeof *this);
  }

  virtual ~ABBA() {
    abba = 0;
    if (is_partial<ABBA,A,B>(this))
      print_layout("ABBA", "dtor_",
		   this, sizeof *this - sizeof(A)-sizeof(B));
    else
      print_layout("ABBA", "dtor", this, sizeof *this);
  }

  ABBA* g () { // overrides AB::g *and* BA::g
    print("g");
    return this;
  }

  ABBA* h() { // overrides AB::h
    print("h");
    return this;
  }

protected:
  long abba;

private:
  void print(const char *func) const {
    std::cout << "ABBA  ::" << func << '('
	      << reinterpret_cast<size_t>(this) << ")\t"
	      << "a = " << a
	      << ", b = " << b
	      << ", ab = " << ab
	      << ", ba = " << ba
	      << ", abba = " << abba << "\n";
  }
};

struct D {
  D(long d_) : d(d_) {
    print_layout("D", "ctor", this, sizeof *this);
  }
  ~D() {
    d = 0;
    print_layout("D", "dtor", this, sizeof *this);
  };

protected:
  long d;
};

struct DCABBA : D, C, ABBA {
  DCABBA(long a, long b, long c, long d,
	 long ab, long ba, long abba, long dcabba_) :
    B(b), A(a), D(d), C(b,c), ABBA(a,b,ab,ba,abba), dcabba(dcabba_) {
    if (is_partial<DCABBA,A,B>(this))
      print_layout("DCABBA", "ctor_",
		   this, sizeof *this - sizeof(A)-sizeof(B));
    else
      print_layout("DCABBA", "ctor", this, sizeof *this);
  }
  virtual ~DCABBA() {
    dcabba = 0;
    if (is_partial<DCABBA,A,B>(this))
      print_layout("DCABBA", "dtor_",
		   this, sizeof *this - sizeof(A)-sizeof(B));
    else
      print_layout("DCABBA", "dtor", this, sizeof *this);
  }
  
  void f() const { // override ABBA::f
    print("f");
  }

  DCABBA* g () { // override ABBA::g and C::g
    print("g");
    return this;
  }

  DCABBA* h() { // overrides ABBA::h
    print("h");
    return this;
  }

protected:
  long dcabba;

private:
  void print(const char *func) const {
    std::cout << "DCABBA::" << func << '('
	      << reinterpret_cast<size_t>(this) << ")\t"
	      << "a = " << a
	      << ", b = " << b
	      << ", c = " << c
	      << ", d = " << d
	      << ", ab = " << ab
	      << ", ba = " << ba
	      << ", abba = " << abba
	      << ", dcabba = " << dcabba << "\n";
  }
};

void dynamic_cast_speed_B_A(B* b)
{
  clock_t t0, t1;
  int i;
  A *a;
  
  t0 = clock();
  for (i = 0; i < 10000000; i++)
    a = dynamic_cast<A*>(b);
  t1 = clock();
    
  std::cout << "dynamic_cast<A*>(b) = "
	    << (size_t)a
	    << ", iter = "
	    << (long)(10000000/((double)t1-t0)*CLOCKS_PER_SEC)
	    << "/sec\n";
}

void dynamic_cast_speed_B_C(B* b)
{
  clock_t t0, t1;
  int i;
  C *c;
  
  t0 = clock();
  for (i = 0; i < 10000000; i++)
    c = dynamic_cast<C*>(b);
  t1 = clock();
    
  std::cout << "dynamic_cast<C*>(b) = "
	    << (size_t)c
	    << ", iter = "
	    << (long)(10000000/((double)t1-t0)*CLOCKS_PER_SEC)
	    << "/sec\n";
}

/*
  ===============================
*/

int main(int argc, char *argv[])
{
  int  do_flags = argc > 1 ? strtol(argv[1],0,0) : 0;
  bool do_layout = do_flags & 1;

  print_layout_flag(do_layout);

  do_layout && std::cout << "*** A Layout\n";
  A a(1);

  do_layout && std::cout << "*** B Layout\n";
  B b(2);

  do_layout && std::cout << "*** C Layout\n";
  C c(2,3);

  do_layout && std::cout << "*** D Layout\n";
  D d(4);

  do_layout && std::cout << "*** AB Layout\n";
  AB ab(1,2,5);

  do_layout && std::cout << "*** BA Layout\n";
  BA ba(1,2,6);

  do_layout && std::cout << "*** ABBA Layout\n";
  ABBA abba(1,2,5,6,7);

  do_layout && std::cout << "*** DCABBA Layout\n";
  DCABBA dcabba(1,2,3,4,5,6,7,8);

  std::cout << "sizeof a      =  " << sizeof a
	    << " @ " << (size_t)&a    << std::endl;
  std::cout << "sizeof b      =  " << sizeof b
	    << " @ " << (size_t)&b    << std::endl;
  std::cout << "sizeof c      = " << sizeof c
	    << " @ " << (size_t)&c    << std::endl;
  std::cout << "sizeof d      = " << sizeof d
	    << " @ " << (size_t)&d    << std::endl;
  std::cout << "sizeof ab     = " << sizeof ab
	    << " @ " << (size_t)&ab   << std::endl;
  std::cout << "sizeof ba     = " << sizeof ba
	    << " @ " << (size_t)&ba   << std::endl;
  std::cout << "sizeof abba   = " << sizeof abba
	    << " @ " << (size_t)&abba << std::endl;
  std::cout << "sizeof dcabba = " << sizeof dcabba
	    << " @ " << (size_t)&dcabba << std::endl;

  a.f();
  b.g();
  c.g();
  static_cast<B*>(&c)->g();

  ab.f();
  ab.g();
  ab.h();
  static_cast<A*>(&ab)->f();
  static_cast<B*>(&ab)->g();
  
  ba.f();
  ba.g();
  static_cast<A*>(&ba)->f();
  static_cast<B*>(&ba)->g();

  abba.f();
  abba.g();
  abba.h();
  ((AB*)&abba)->f();             // bypass protection
  ((AB*)&abba)->g();             // bypass protection
  ((AB*)&abba)->h();             // bypass protection
  static_cast<BA*>(&abba)->f();
  static_cast<BA*>(&abba)->g();
  static_cast<A*>(&abba)->f();
  static_cast<B*>(&abba)->g();

  dcabba.f();
  dcabba.g();
  dcabba.h();
  static_cast<C*>(&dcabba)->g();
  static_cast<ABBA*>(&dcabba)->f();
  static_cast<ABBA*>(&dcabba)->g();
  static_cast<ABBA*>(&dcabba)->h();
  ((AB*)&dcabba)->f();          // bypass protection
  ((AB*)&dcabba)->g();          // bypass protection
  ((AB*)&dcabba)->h();          // bypass protection
  static_cast<BA*>(&dcabba)->f();
  static_cast<BA*>(&dcabba)->g();
  static_cast<A*>(&dcabba)->f();
  static_cast<B*>(&dcabba)->g();

  std::cout << "AB    : ";
  dynamic_cast_speed_B_A(&ab);
  std::cout << "DCABBA: ";
  dynamic_cast_speed_B_A(&dcabba);
  std::cout << "DCABBA: ";
  dynamic_cast_speed_B_C(&dcabba);

  do_layout && std::cout << "*** DCABBA Layout\n";
  dcabba.~DCABBA();

  do_layout && std::cout << "*** ABBA Layout\n";
  abba.~ABBA();

  do_layout && std::cout << "*** BA Layout\n";
  ba.~BA();

  do_layout && std::cout << "*** AB Layout\n";
  ab.~AB();

  do_layout && std::cout << "*** D Layout\n";
  d.~D();

  do_layout && std::cout << "*** C Layout\n";
  c.~C();

  do_layout && std::cout << "*** B Layout\n";
  b.~B();

  do_layout && std::cout << "*** A Layout\n";
  a.~A();

  print_layout_flag(false);

  // Objects will be destroyed twice (with no side effects) for the
  // purpose of the layout display. This not a recommended style !
}
