#ifndef OBJECT_MODEL_H
#define OBJECT_MODEL_H

#include <stddef.h>

/* -----
 * NOTE: Virtual table layout
 * -----
 *
 * offset   content
 * -----------------------------------
 * -N       Nth vbase offset
 * ...      ...
 * -3       1st vbase offset
 * -2       offset to top
 * -1       pointer to type
 *  0       1st virtual function address
 *  1       2nd virtual function address
 * ...
 *
 * This layout assumes that pointers to data and pointers to function
 * fit into a long. If it not the case, the compiler is free to use
 * internally the appropriate integral type.
 *
 * -----
 */

typedef ptrdiff_t VTABLE;

#define VCALL0(T,M) \
        ((TYPE(M)*)VFUNC(T,M))(T)

#define VCALL1(T,M,A1) \
        ((TYPE(M)*)VFUNC(T,M))(T,A1)

#define VCALL2(T,M,A1,A2) \
        ((TYPE(M)*)VFUNC(T,M))(T,A1,A2)

#define VFUNC(T,M) \
        ((T)->vptr[INDEX(M)])

#define VBASE(D,VB) \
        ((VB*)((char*)(D) + (D)->vptr[INDEX(VB)]))

#define VTOP(T) \
        ((void*)((char*)(T) + (T)->vptr[INDEX(TOP)]))

#define VOFFSET(D,B) \
        offsetof(D, B ## _vptr)

#define VOFFSET2(D2,D1,B) \
        (VOFFSET(D2,B) - VOFFSET(D2,D1))

#define VSTART(T) \
        (-INDEX(T ## _LAST))

#define BASE(T,B) \
        ((B*)&(T)->B ## _vptr)

#define DERIVED(T,B,D) \
        ((D*)((char*)(T) - VOFFSET(D,B)))

#define NBASE(T) \
        (sizeof(T ## _base)/sizeof(T ## _base[0]))

/* -----
 * NOTE: Virtual functions types and indexes
 * -----
 *
 * vbases index_of are known by the compiler for each class
 * vfunc  index_of are known by the compiler for each class
 * vfunc  type_of  are known by the compiler for each class
 *
 * -----
 */

#define TYPE(T)          type_of_ ## T
#define INDEX(T)        index_of_ ## T
#define AFTER_VFUNC(M)  (INDEX(M )+1)
#define AFTER_VBASE(VB) (INDEX(VB)-1)
#define AFTER_VCALL(VC) (INDEX(VC)-1)

enum {
  FIRST_VFUNC =  0,
  FIRST_VBASE = -3,
  FIRST_VCALL = -2,

  index_of_NO_VFUNC      =  0,
  index_of_NO_VBASE      = -2,
  index_of_NO_VBASE_LAST = -2,

  index_of_TINFO         = -1,
  index_of_TOP           = -2
};

/*
 * Special function replacement for slots of pure virtual function
 */
void pure_virtual_call(void);

/*
 * -----
 * NOTE: Type info
 * -----
 */

typedef struct {
  const char               *name;
  const struct base_info   *base;
  unsigned                  base_cnt;
} type_info;
         
struct base_info {
  const type_info          *base_type;
  const ptrdiff_t           base_offset;
  const enum {
    public_base    = 0x01,
    protected_base = 0x02,
    private_base   = 0x04,
    virtual_base   = 0x08 } base_flag;
};

void*            dynamic_cast(void *this, const type_info *target_type);
const type_info* typeid      (void *this);

#define dynamic_cast(T,B) \
        ((B*)dynamic_cast(&(T)->vptr, B ## _info))

#define typeid(T) \
        typeid(&(T)->vptr)

/*
 * Unused: cancel compiler warning for unused argument.
 */
         
#define UNUSED1(a1)    while(0) {  (void)(a1); }
#define UNUSED2(a1,a2) while(0) { UNUSED1(a1); UNUSED1(a2); }

/*
 * Print utilities
 */

void print_layout_flag(int do_it);
void print_thunk_flag (int do_it);

void print_layout(const char *cls, const char *func,
                  void *this, size_t size);
void print_thunk (const char *cls, const char *func,
                  void *this, int offset, int is_ret);
void print_vfunc (const char *cls, const char *func, const void *this,
                  const char *fmt, ...);
void print_mfunc (const char *cls, const char *func, const void *this,
                  const char *fmt, ...);

#endif
